# Copyright 2008 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""The empty module is here so that pylint can examine the scripts

   @author Robert Crida <robert.crida@ska.ac.za>
   @date 2008-10-10
   """

# Automatically added by katversion
__version__ = '0.9.0'
